/**
 * Noiz
 * (c) Kenta Cho (cs8k-cyu@asahi-net.or.jp)
 * http://www.asahi-net.or.jp/~cs8k-cyu/
 *
 * displaymanager.c(handle screen buffer)
 */
#include "numbers.h"
#include "level.h"
#include "miss.h"
#include "title.h"
#include "now.h"
#include "newgame.h"
#include "continue.h"
#include "quit.h"
#include "arrow.h"

#define SCREENW 27
#define SCREENH 18

#define BITMAP_NUMBER_IDX 486
#define BITMAP_LEVEL_IDX 496
#define BITMAP_MISS_IDX 500
#define BITMAP_SPACE_IDX 504

#define BITMAP_TITLE_IDX 0
#define BITMAP_NOW_IDX 100
#define BITMAP_NEWGAME_IDX 110
#define BITMAP_CONTINUE_IDX 120
#define BITMAP_QUIT_IDX 130
#define BITMAP_ARROW_IDX 140

static void setBitmaps() {
  font_set_monodata(BITMAP_NUMBER_IDX, 10, bmp_numbers);
  font_set_monodata(BITMAP_LEVEL_IDX, level_height, bmp_level);
  font_set_monodata(BITMAP_MISS_IDX, miss_height, bmp_miss);
  font_set_monodata(BITMAP_SPACE_IDX, 1, bmp_space);
}

static void setTitleBitmaps() {
  font_set_monodata(BITMAP_TITLE_IDX, title_width*title_height, bmp_title);
  font_set_monodata(BITMAP_NOW_IDX, now_height, bmp_now);
  font_set_monodata(BITMAP_NEWGAME_IDX, newgame_height, bmp_newgame);
  font_set_monodata(BITMAP_CONTINUE_IDX, continue_height, bmp_continue);
  font_set_monodata(BITMAP_QUIT_IDX, quit_height, bmp_quit);
  font_set_monodata(BITMAP_ARROW_IDX, arrow_height, bmp_arrow1);
  font_set_monodata(BITMAP_ARROW_IDX+1, arrow_height, bmp_arrow2);
  font_set_monodata(BITMAP_ARROW_IDX+2, arrow_height, bmp_arrow3);
}

static void initDisplay() {
  display_control(0x0001);
}

static void clearScreen() {
  int i, j;
  static int data;
  data = BITMAP_SPACE_IDX;
  for ( i=0 ; i<SCREENH ; i++ ) {
    for ( j=0 ; j<SCREENW+1 ; j++ ) {
      screen_set_char(SCREEN1, j, i, 1, 1, &data);
    }
  }
}

static void drawBitmapWd(int idx, int x, int y, int width, int height) {
  static int data;
  int i, j;
  data = idx;
  for ( i=0 ; i<height ; i++ ) {
    for ( j=0 ; j<width ; j++ ) {
      screen_set_char(SCREEN1, j+x, i+y, 1, 1, &data);
      data++;
    }
  }
}

static void drawBitmap(int idx, int x, int y, int height) {
  static int data;
  int i;
  data = idx;
  for ( i=0 ; i<height ; i++ ) {
    screen_set_char(SCREEN1, x, i+y, 1, 1, &data);
    data++;
  }
}

static void clearBitmap(int x, int y, int height) {
  static int data;
  int i;
  data = BITMAP_SPACE_IDX;
  for ( i=0 ; i<height ; i++ ) {
    screen_set_char(SCREEN1, x, y+i, 1, 1, &data);
  }
}

unsigned char buffer[SCREENW * SCREENH * 8];

static void clearDisplay() {
  memset(buffer, 0, sizeof(buffer));
}

static void flipDisplay() {
  font_set_monodata(0, SCREENW*SCREENH, buffer);
}

/**
 * set characters to screen
 */
static void createScreenBuffer() {
  static int data;
  int i;
  clearDisplay();
  flipDisplay();
  for ( i=0 ; i<SCREENW*SCREENH ; i++ ) {
    data = i;
    screen_set_char(SCREEN1, i%SCREENW, i/SCREENW, 1, 1, &data);
  }
}

#define SCREEN_WIDTH 216
#define SCREEN_HEIGHT 144

/*
 * this function moved to putline.a86
 */
/**
 * draw line to the character generator
 */
/*static void WinDrawLine(int x1, int y1, int x2, int y2) {
  int a, b;
  int i, px, py;

  if ( x1 < 0 ) {
    if ( x2 < 0 ) return;
    y1 = ((long)y1-y2)*x2/(x2-x1)+y2;
    x1 = 0;
  } else if ( x2 < 0 ) {
    if ( x1 < 0 ) return;
    y2 = ((long)y2-y1)*x1/(x1-x2)+y1;
    x2 = 0;
  }
  if ( x1 >= SCREEN_WIDTH ) {
    if ( x2 >= SCREEN_WIDTH ) return;
    y1 = ((long)y1-y2)*(x2-(SCREEN_WIDTH-1))/(x2-x1)+y2;
    x1 = SCREEN_WIDTH-1;
  } else if ( x2 >= SCREEN_WIDTH ) {
    if ( x1 >= SCREEN_WIDTH ) return;
    y2 = ((long)y2-y1)*(x1-(SCREEN_WIDTH-1))/(x1-x2)+y1;
    x2 = SCREEN_WIDTH-1;
  }
  if ( y1 < 0 ) {
    if ( y2 < 0 ) return;
    x1 = ((long)x1-x2)*(y2-1)/(y2-y1)+x2;
    y1 = 0;
  } else if ( y2 < 0 ) {
    if ( y1 < 0 ) return;
    x2 = ((long)x2-x1)*(y1-1)/(y1-y2)+x1;
    y2 = 0;
  }
  if ( y1 >= SCREEN_HEIGHT ) {
    if ( y2 >= SCREEN_HEIGHT ) return;
    x1 = (long)(x1-x2)*(y2-(SCREEN_HEIGHT-1))/(y2-y1)+x2;
    y1 = SCREEN_HEIGHT-1;
  } else if ( y2 >= SCREEN_HEIGHT ) {
    if ( y1 >= SCREEN_HEIGHT ) return;
    x2 = (long)(x2-x1)*(y1-(SCREEN_HEIGHT-1))/(y1-y2)+x1;
    y2 = SCREEN_HEIGHT-1;
  }

  a = abs(x2 - x1);
  b = abs(y2 - y1);
  l = max(a, b);
  ax = (((long)x2 - x1)<<8) / l;
  ay = (((long)y2 - y1)<<8) / l;
  x = x1<<8;
  y = y1<<8;
  for ( i=0 ; i<l ; i++, x+=ax, y+=ay ) {
   py = y>>8;
    px = x>>8;
    buffer[(px&~7) + (py&7) + (py&~7) * SCREENW] |= 0xc0>>(px&7);
  }
}*/


